@@ -6,8 +6,6 @@ frames as specified in `sections 4 to 8 of RFC 6455`_.
 
 """
 
-__all__ = ['WebSocketCommonProtocol']
-
 import asyncio
 import asyncio.queues
 import codecs
@@ -23,6 +21,8 @@ from .framing import *
 from .handshake import *
 
 
+__all__ = ['WebSocketCommonProtocol']
+
 logger = logging.getLogger(__name__)
 
 
@@ -62,6 +62,22 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
     raise :exc:`~websockets.exceptions.ConnectionClosed` and the connection
     will be closed with status code 1009.
 
+    The ``max_queue`` parameter sets the maximum length of the queue that holds
+    incoming messages. The default value is 32. 0 disables the limit. Messages
+    are added to an in-memory queue when they're received; then :meth:`recv()`
+    pops from that queue. In order to prevent excessive memory consumption when
+    messages are received faster than they can be processed, the queue must be
+    bounded. If the queue fills up, the protocol stops processing incoming data
+    until :meth:`recv()` is called. In this situation, various receive buffers
+    (at least in ``asyncio`` and in the OS) will fill up, then the TCP receive
+    window will shrink, slowing down transmission to avoid packet loss.
+
+    Since Python can use up to 4 bytes of memory to represent a single
+    character, each websocket connection may use up to ``4 * max_size *
+    max_queue`` bytes of memory to store incoming messages. By default,
+    this is 128MB. You may want to lower the limits, depending on your
+    application's requirements.
+
     Once the handshake is complete, request and response HTTP headers are
     available:
 
@@ -87,8 +103,8 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
 
     def __init__(self, *,
                  host=None, port=None, secure=None,
-                 timeout=10, max_size=2 ** 20, loop=None,
-                 legacy_recv=False):
+                 timeout=10, max_size=2 ** 20, max_queue=2 ** 5,
+                 loop=None, legacy_recv=False):
         self.host = host
         self.port = port
         self.secure = secure
@@ -97,6 +113,8 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         self.max_size = max_size
         # Store a reference to loop to avoid relying on self._loop, a private
         # attribute of StreamReaderProtocol, inherited from FlowControlMixin.
+        if loop is None:
+            loop = asyncio.get_event_loop()
         self.loop = loop
 
         self.legacy_recv = legacy_recv
@@ -128,13 +146,13 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         self.connection_closed = asyncio.Future(loop=loop)
 
         # Queue of received messages.
-        self.messages = asyncio.queues.Queue(loop=loop)
+        self.messages = asyncio.queues.Queue(max_queue, loop=loop)
 
         # Mapping of ping IDs to waiters, in chronological order.
         self.pings = collections.OrderedDict()
 
         # Task managing the connection, initalized in self.client_connected.
-        self.worker = None
+        self.worker_task = None
 
         # In a subclass implementing the opening handshake, the state will be
         # CONNECTING at this point.
@@ -220,12 +238,12 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         # the worker loop.
         try:
             yield from asyncio.wait_for(
-                self.worker, self.timeout, loop=self.loop)
+                self.worker_task, self.timeout, loop=self.loop)
         except asyncio.TimeoutError:
-            self.worker.cancel()
+            self.worker_task.cancel()
 
         # The worker should terminate quickly once it has been cancelled.
-        yield from self.worker
+        yield from self.worker_task
 
     @asyncio.coroutine
     def recv(self):
@@ -262,7 +280,7 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
             self.messages.get(), loop=self.loop)
         try:
             done, pending = yield from asyncio.wait(
-                [next_message, self.worker],
+                [next_message, self.worker_task],
                 loop=self.loop, return_when=asyncio.FIRST_COMPLETED)
         except asyncio.CancelledError:
             # Handle the Task.cancel()
@@ -379,7 +397,7 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         # longer than the worst case (2 * self.timeout) but not unlimited.
         if self.state == CLOSING:
             yield from asyncio.wait_for(
-                self.worker, 3 * self.timeout, loop=self.loop)
+                self.worker_task, 3 * self.timeout, loop=self.loop)
             raise ConnectionClosed(self.close_code, self.close_reason)
 
         # Control may only reach this point in buggy third-party subclasses.
@@ -395,7 +413,7 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
                 msg = yield from self.read_message()
                 if msg is None:
                     break
-                self.messages.put_nowait(msg)
+                yield from self.messages.put(msg)
             except asyncio.CancelledError:
                 break
             except WebSocketProtocolError:
@@ -612,7 +630,7 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
         self.reader = reader
         self.writer = writer
         # Start the task that handles incoming messages.
-        self.worker = asyncio_ensure_future(self.run(), loop=self.loop)
+        self.worker_task = asyncio_ensure_future(self.run(), loop=self.loop)
 
     def eof_received(self):
         super().eof_received()
@@ -631,6 +649,8 @@ class WebSocketCommonProtocol(asyncio.StreamReaderProtocol):
     def connection_lost(self, exc):
         # 7.1.4. The WebSocket Connection is Closed
         self.state = CLOSED
+        if not self.opening_handshake.done():
+            self.opening_handshake.set_result(False)
         if not self.closing_handshake.done():
             self.close_code, self.close_reason = 1006, ''
             self.closing_handshake.set_result(False)
